
#include "FINDER.h"


/*
*	Do_smallrefs is called after the search space has been prepared, to 
*	record refutations known to be of cardinality setting.pre_test or less.
*	Injective and surjective functions are also treated at this stage. 
*
*	The first bit of this function sets the matrix values of the nulladic 
*	functions which are fixed and which therefore don't count in refutations.
*/

int do_smallrefs( unsigned info[], TRS *T )
{
  int twt, i;
  clause c;
  functor f;
  wff plit;

  for ( f = FN; f; f = f->nextfunct ) if ( !f->adicity )
    if ( !(info[f->fcc] & (info[f->fcc]-1)) ) {   // is this 'sposed to be &&
      for ( i = 0; !(info[f->fcc] & (1<<i)); i++ );
      *(f->table) = i;
    }

  do_injectives( T );
  if ( !do_surjectives(T) ) return 0;

  for ( twt = 0; twt <= setting.pretest_limit; twt++ ) {
    for ( c = TH[current_theory]; c; c = c->nextclause ) 
      if ( c->weight > 1 )
	if ( !T->cut_flag )
	  if ( localweight(info,c,&plit,T) == twt )
	    if ( !twt_test(info, c, twt, plit, T) )
	      return 0;
  }
  return 1;
}



/*
*	The test routine for a clause whose local weight is lwt.
*	Return value is false iff a 0-refutation has been found.
*/

boolean twt_test( unsigned info[], clause c, int lwt, wff plit, TRS *T )
{
  wff_stack wf, tf[PRE_MAX];
  int maximum[PRE_MAX];
  int i;
  functor f;

#ifdef DEBUG
  print_one_clause(stdout,c);
  fflush(stdout);
#endif

  tf[lwt] = 0;
  maximum[lwt] = 0;
  if ( lwt ) tf[0] = firstopen( info, c->testlist, lwt );
  for ( i = 1; i < lwt; i++ )
    tf[i] = firstopen( info, tf[i-1]->nxt, 1 );
  for ( i = 0; i < lwt; i++ )
    maximum[i] = tf[i]->w->sym->val_type->siz+1;
  correct_order( tf );

/*
*	First put in the values of any fixed cells
*/
  for ( wf = c->testlist; wf; wf = wf->nxt ) {
    f = wf->w->sym;
    if ( is_open(f) ) {
      if ( !f->adicity && info[f->fcc] &&
	   !(info[f->fcc] & (info[f->fcc]-1)) )
	for ( i = 0; !(info[f->fcc] & (1 << i)); i++ ) ;
      *(f->table) = i;
    }
  }

/*
*	Next initialize all the variables to the first (highest) value.
*/
  for ( wf = c->varlist; wf; wf = wf->nxt )
    wf->w->val = wf->w->sym->val_type->siz-1;
/*
*	Now repeat the test until no more variable assignments remain.
*/
  do if ( !alive(c, tf, tf, maximum, c->testlist, info, plit, T) )
    return false;
  while ( !T->cut_flag && next_assignment(c) );
  return true;
}



boolean alive(clause c, wff_stack *tf, wff_stack *ltf, int *maximum,
	      wff_stack wf, unsigned info[], wff plit, TRS *T)
{
  int i;
  int subvalues[ARGMAX];

  for ( ; wf != *ltf; wf = wf->nxt ) {
    for (i=0; i<wf->w->sym->adicity; i++)
      subvalues[i] = *(wf->w->sbval[i]);
    wf->w->val = *(wf->w->matrix + getoffset(wf->w->sym,subvalues));
  }

  if ( !*ltf ) return good_all( c, tf, plit, T );

  if ( wf->w == plit )
    return (alive(c, tf, ltf+1, maximum+1, (*ltf)->nxt, info, plit, T));

  for ( (*ltf)->w->val = 0; (*ltf)->w->val < *maximum; (*ltf)->w->val++ )
    if ( poss_val((*ltf)->w,info) )
      if ( !alive(c, tf, ltf+1, maximum+1, (*ltf)->nxt, info, plit, T) )
	return false;
  return true;
}




/*
*	If the clause fails, inform TRANSREF. Return false iff the result
*	refutes the entire search space (this will cause exit from do_twos).
*/

boolean good_all( clause c, wff_stack tf[], wff plit, TRS *T)
{
  boolean failure = true;
  wff_stack wf;
  wff W;
  int cells[PRE_MAX], vals[PRE_MAX];
  boolean valency[PRE_MAX];
  int i,j;
  int subvalues[ARGMAX];

  for ( wf = c->prem; wf; wf = wf->nxt )
    if ( !wf->w->val ) failure = false;
  if ( plit ) plit->val = c->conc->w->sb[1]->val;
  else for ( wf = c->conc; wf; wf = wf->nxt )
    if ( wf->w->val ) failure = false;

  if ( failure ) {
    for ( i = 0; tf[i]; i++ ) {
      W = tf[i]->w;
      for (j=0; j<W->sym->adicity; j++)
	subvalues[j] = *(W->sbval[j]);
      cells[i] = vec_pos(W->sym, subvalues);
      vals[i] = W->val;
      valency[i] = (W == plit);
    }
    cells[i] = -1;
    new_small_ref( cells, vals, valency, T );
    if ( T->novect ) return false;
  }
  return true;
}




/*
*	Poss_val returns true iff the value assigned to w is possible
*	according to the vector of possible values. Note that this 
*	calls function vec_pos.
*
*	If one of the input values is the dummy, then poss_val is true 
*	iff the value of w is the dummy of the appropriate type.
*/

boolean poss_val( wff w, unsigned info[] )
{
  int i;
  int args[ARGMAX];

  for (i=0; i<w->sym->adicity; i++)
    if (*(w->sbval[i]) == w->sym->arg_type[i]->siz)
      return (w->val==w->sym->val_type->siz);
  for (i=0; i<w->sym->adicity; i++)
    args[i] = *(w->sbval[i]);
  return !!(info[vec_pos(w->sym,args)] & (1 << w->val));
}



/*
*	For testing purposes, tfb is assumed to be in the posterity
*	of the other open cell tfa. They just might have got assigned
*	the wrong way round, so this function tests for correctness.
*/

void correct_order( wff_stack tf[] )
{
    wff_stack wf;
    int i;

    if ( !tf[0] || !tf[1] ) return;
    for ( i = 0; tf[i+1]; i++ ) {
	for ( wf = tf[i]->nxt; wf; wf = wf->nxt )
	    if ( wf == tf[i+1] ) break;
	if ( !wf ) skipout("Bad pointer sequence in a clause",1);
    }
}




/*
*	Injectiveness is special to TRANSREF. It has to be informed of
*	the set of cells whose values cannot be identical, of the upper
*	bound on the set of values which cannot be assigned to more 
*	than one of the cells, and of course of the TRS structure to 
*	which all of this pertains. The upper bound on values for this 
*	purpose is to allow partial functions to be injective.
*/

void do_injectives( TRS *T )
{
    unsigned dinfo[V_LENGTH];
    int i, j;
    functor f;
    int argument[ARGMAX];
		
    for ( f = FN; f; f = f->nextfunct ) if ( !f->piped ) {
	if ( f->injective ) {
	    for ( i = 0; i < tr_parameters.vlength; i++ )
		dinfo[i] = 0;
	    set_dinfo(f,f->adicity-1,argument,dinfo);
	    injective( dinfo, f->val_type->siz, T );
	    f->leftin = f->rightin = false;
	}
	if ( f->rightin ) {
	    for ( i = 0; i < tr_parameters.vlength; i++ )
		dinfo[i] = 0;
	    for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
		argument[0] = i;
		for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
		    argument[1] = j;
		    dinfo[vec_pos(f,argument)] = 1;
		}
		injective( dinfo, f->val_type->siz, T );
		for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
		    argument[1] = j;
		    dinfo[vec_pos(f,argument)] = 0;
		}
	    }
	}
	if ( f->leftin ) {
	    for ( i = 0; i < tr_parameters.vlength; i++ )
		dinfo[i] = 0;
	    for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
		argument[1] = j;
		for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
		    argument[0] = i;
		    dinfo[vec_pos(f,argument)] = 1;
		}
		injective( dinfo, f->val_type->siz, T );
		for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
		    argument[0] = i;
		    dinfo[vec_pos(f,argument)] = 0;
		}
	    }
	}
    }
}



void set_dinfo(functor f, int x, int argument[], unsigned dinfo[])
{
    for (argument[x]=0; argument[x]<f->arg_type[x]->siz; argument[x]++) {
	if (x)
	    set_dinfo(f,x-1,argument,dinfo);
	else
	    dinfo[vec_pos(f,argument)] = 1;
    }
}




/*
*	Surjective functions are also treated specially by TRANSREF.
*	Left and right surjectiveness are treated here, as well as 
*	total surjectiveness.
*/

boolean do_surjectives( TRS *T )
{
    unsigned dinfo[V_LENGTH], vset;
    int i, j, m;
    int valcount, argcount;
    functor f;
    int argument[ARGMAX];

    for ( f = FN; f; f = f->nextfunct ) if ( !f->piped ) {
	if ( f->rightsur ) {
	    for ( i = 0; i < tr_parameters.vlength; i++ ) dinfo[i] = 0;
	    for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
		argument[0] = i;
		for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
		    argument[1] = j;
		    dinfo[vec_pos(f,argument)] = 1;
		}
		vset = 0;
		valcount = 0;
		for ( m = 0; m < f->val_type->siz; m++ ) {
		    vset |= (1 << m);
		    valcount++;
		}
		if ( valcount > f->arg_type[1]->siz ) return false;
		surjective( dinfo, vset, T );
		for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
		    argument[1] = j;
		    dinfo[vec_pos(f,argument)] = 0;
		}
	    }
	    f->surjective = false;
	}
	if ( f->leftsur ) {
	    for ( i = 0; i < tr_parameters.vlength; i++ ) dinfo[i] = 0;
	    for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
		argument[1] = j;
		for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
		    argument[0] = i;
		    dinfo[vec_pos(f,argument)] = 1;
		}
		vset = 0;
		valcount = 0;
		for ( m = 0; m < f->val_type->siz; m++ ) {
		    vset |= (1 << m);
		    valcount++;
		}
		if ( valcount > f->arg_type[0]->siz ) return false;
		surjective( dinfo, vset, T );
		for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
		    argument[0] = i;
		    dinfo[vec_pos(f,argument)] = 0;
		}
		f->surjective = false;
	    }
	}
	if ( f->surjective ) {
	    for ( i = 0; i < tr_parameters.vlength; i++ ) dinfo[i] = 0;
	    set_dinfo(f,f->adicity-1,argument,dinfo);
	    argcount = 1;
	    for (i=0; i<f->adicity; i++)
		argcount *= f->arg_type[i]->siz;
	    vset = 0;
	    valcount = 0;
	    for ( i = 0; i < f->val_type->siz; i++ ) {
		vset |= (1 << i);
		valcount++;
	    }
	    if ( valcount > argcount ) return false;
	    surjective( dinfo, vset, T );
	}
    }
    return true;
}



/*
*	Local weight is weight minus any nulladic functions with only
*	one possible value. The parameter *plit is set to point to the
*	left subformula of the sole positive literal (if any) which must
*	have identity as its predicate and an open left subformula.
*/ 

int localweight( unsigned info[], clause c, wff *plit, TRS *T )
{
    wff_stack wf;
    int counter = 0;
    functor f;
    boolean poscon;		/* Positive literal condition */

    poscon = (
	      c->conc &&
	      !c->conc->nxt &&
	      *(c->conc->w->sym->fname) == '=' &&
	      is_open(c->conc->w->sb[0]->sym) &&
	      !occurs_inside(c->conc->w->sb[0], c->conc->w->sb[1]) &&
	      !occurs_in_list(c->conc->w->sb[0], c->prem) );
    if ( poscon ) *plit = c->conc->w->sb[0];
    else *plit = 0;

    for ( wf = c->testlist; wf; wf = wf->nxt ) {
	f = wf->w->sym;
	if ( is_open(f) ) {
	    if ( f->adicity || ((info[f->fcc] & (info[f->fcc]-1))) )
		counter++;
	}
    }
    return counter;
}



/*
*	The first open spot in a list of wffs is the first one which
*	is part of a function that is_open, other than a nulladic one
*	with only one possible value.
*/

wff_stack firstopen( unsigned info[], wff_stack wfl, int opens )
{
    functor f;

    f = wfl->w->sym;
    while ( !is_open(f) ||
	   (opens && !(f->adicity || (info[f->fcc] & (info[f->fcc]-1)))) ) {
	wfl = wfl->nxt;
	f = wfl->w->sym;
    }
    return wfl;
}



/*
*	A function is_open if it is really part of the search space 
*	and not one of the constant functions such as identity or a
*	function piped in from another FINDER.
*/

boolean is_open( functor f )
{
    if ( !f ) return false;
    if ( f->category != Func && f->category != Const ) return false;
    if ( strchr("+=<>",f->fname[0]) ) return false;
    if ( f->fname[0] == 'E' && f->fname[1] == '!' ) return false;
    if ( f->piped ) return false;
    return true;
}


/*
*	Formula fmla occurs as a subformula of something in wfl.
*/

boolean occurs_in_list( wff fmla, wff_stack wfl )
{
    if ( !wfl ) return false;
    do
	if ( occurs_inside(fmla,wfl->w) ) return true;
    while ( (wfl = wfl->nxt) );
    return false;
}


/*
*	Formula little occurs as a subformula of formula big.
*/

boolean occurs_inside( wff little, wff big )
{
    int i;;

    if ( !big ) return false;
    if ( little == big ) return true;
    for (i=0; i<big->sym->adicity; i++)
	if ( occurs_inside(little, big->sb[i]) ) return true;
    return false;
}
