/*-------------------------------------------------------------------------*\
  The following are copies of OTTER macros which are often undefined.
  However SEMANTIC GUIDANCE still requires them
\*-------------------------------------------------------------------------*/


/*-------------------------------------------------------------------------*\
  SCOTT_CPU_TIME(sec, usec)
  It has been sec seconds + usec microseconds since the start of this process.
\*-------------------------------------------------------------------------*/

#ifdef TURBO_C  /* DOS */
#define SCOTT_CPU_TIME(sec, usec) \
{ \
    struct timeb r; \
    ftime(&r); \
    sec = r.time; \
    usec = (long) r.millitm * 1000; \
}/*--SCOTT_CPU_TIME()---------------------------------------------------*/
#else

#ifdef THINK_C  /* Macintosh */
#define SCOTT_CPU_TIME(sec, usec) \
{ \
    long ticks; \
    ticks = clock(); \
    sec = ticks / CLOCKS_PER_SEC; \
    usec = (ticks % CLOCKS_PER_SEC) * (1000000 / CLOCKS_PER_SEC); \
}/*--SCOTT_CPU_TIME()---------------------------------------------------*/
#else

#ifdef TP_RUSAGE
#define SCOTT_CPU_TIME(sec, usec) \
{ \
    struct rusage r; \
    getrusage(RUSAGE_SELF, &r); \
    sec = r.ru_utime.tv_sec; \
    usec = r.ru_utime.tv_usec; \
}/*--SCOTT_CPU_TIME()---------------------------------------------------*/

#else
#define SCOTT_CPU_TIME(sec, usec) {sec = usec = 0;}
#endif // TURBO_C
#endif // Macintosh
#endif // TP_RUSAGE

/*----------------------------------------------------------------------*\
  SCOTT_CLOCK_START(clock_num)

  Start or continue timing SEMANTIC GUIDANCE.  If the clock is already
  running, a warning message is printed.
\*----------------------------------------------------------------------*/
#define SCOTT_CLOCK_START(c) \
{ \
    struct clock *cp; \
 \
    cp = &Clocks[c]; \
    if (cp->curr_sec != -1) { \
	fprintf(stderr, "WARNING, CLOCK_START: clock %d already on.\n", c); \
	printf("WARNING, CLOCK_START: clock %d already on.\n", c); \
	} \
    else \
	CPU_TIME(cp->curr_sec, cp->curr_usec) \
}/*--CLOCK_START()------------------------------------------------------*/


/*----------------------------------------------------------------------*\
  SCOTT_CLOCK_STOP(clock_num)

  Stop timing clock and add to accumulated total.  If the clock not
  running, a warning message is printed.
\*----------------------------------------------------------------------*/
#define SCOTT_CLOCK_STOP(c) \
{ \
    long sec, usec; \
    struct clock *cp; \
 \
    cp = &Clocks[c]; \
    if (cp->curr_sec == -1) { \
	fprintf(stderr, "WARNING, SCOTT_CLOCK_STOP: clock %d already off.\n", c); \
	printf("WARNING, SCOTT_CLOCK_STOP: clock %d already off.\n", c); \
	} \
    else { \
	CPU_TIME(sec, usec) \
	cp->accum_sec += sec - cp->curr_sec; \
	cp->accum_usec += usec - cp->curr_usec; \
	cp->curr_sec = -1; \
	cp->curr_usec = -1; \
	} \
}/*--CLOCK_STOP()------------------------------------------------------*/

