#include "SCOTT.h"

static struct data_array cost_of_given; 
static struct data_array cost_per_clause; 
static struct data_array time_saved;
static struct data_array change_in_time_saved;
static struct data_array prob_of_success;

void init_costs()
  /*------------------------------------------------------------------
   * Called by read_all_input()
   * Initialise data structures associated with this file.
   * -----------------------------------------------------------------*/
{
  int point;

  strcpy(cost_of_given.name,"cost_of_given");
  cost_of_given.num_points = 0;
  cost_of_given.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    cost_of_given.list[point] = 0; 
  
  strcpy(cost_per_clause.name,"cost_per_clause");
  cost_per_clause.num_points = 0;
  cost_per_clause.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    cost_per_clause.list[point] = 0;
  
  strcpy(time_saved.name,"time_saved");
  time_saved.num_points = 0;
  time_saved.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    time_saved.list[point] = 0;

  strcpy(change_in_time_saved.name,"change_in_time_saved");
  change_in_time_saved.num_points = 0;
  change_in_time_saved.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    change_in_time_saved.list[point] = 0;
  
  strcpy(prob_of_success.name,"prob_of_success");
  prob_of_success.num_points = 0;
  prob_of_success.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    prob_of_success.list[point] = 0;
  
}/*init_costs()-------------------------------------------------------*/

void reset_prob_of_success()
  /*------------------------------------------------------------------
   * Called by execute_strategy().
   * 
   * Time limit has permanately changed so discard previous estimates
   * of extention success.
   * -----------------------------------------------------------------*/
{
  int point;

  prob_of_success.num_points = 0;
  prob_of_success.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    prob_of_success.list[point] = 0;
}

void set_data_array(struct data_array *array, float value)  
  /*------------------------------------------------------------------
   * Called by various functions

   * Append "value" to a finite list of data points.  If we have reach
   * our limit on data point (NUM_POINTS), pop off the first
   * value in the list before appending the new value.  Note we assume
   * all values are positive integers.
   * -----------------------------------------------------------------*/
{
  int point;

  if (array->num_points < NUM_POINTS) {
    array->list[array->num_points] = value;
  }
  else { // shuffle along
    for (point = 1; point < NUM_POINTS; point++)
      array->list[point - 1] = array->list[point];
    array->list[point - 1] = value;
  }
  
  array->num_points++;
  array->total += value;
  
}/*set_data_array()--------------------------------------------------*/


void set_cost_of_given(float value)
  /*------------------------------------------------------------------
   * Called by semantic_selection().  Set next cost_of_given.
   * -----------------------------------------------------------------*/
{
  set_data_array(&cost_of_given, value);

}/*set_cost_of_given()------------------------------------------------*/


void set_cost_per_clause(float value)
  /*------------------------------------------------------------------
   * Called by execute_strategy().  Set next cost_per_clause.
   * -----------------------------------------------------------------*/
{
  set_data_array(&cost_per_clause, value);

}/*set_cost_per_clause()----------------------------------------------*/

void set_prob_of_success(float value)
  /*------------------------------------------------------------------
   * Called by execute_strategy().  Set next "prob_of_success"
   * -----------------------------------------------------------------*/
{
  set_data_array(&prob_of_success,value);
  
}/*set_cost_per_clause()----------------------------------------------*/


void set_time_saved(int givens_avoided)
  /*------------------------------------------------------------------
   * Called by persist().  Set the nect "time saved" and if that array
   * is full, set the next "change_in_time_saved";
   * -----------------------------------------------------------------*/
{
  static int last_givens_avoided = 0;
  
  if (givens_avoided){
    float value;
    
    value = est_from_data_array(&cost_of_given)*
      (givens_avoided - last_givens_avoided)*rate_partition();

    set_data_array(&time_saved,value);
    
    if (time_saved.num_points >= NUM_POINTS) // list is full
      set_change_in_time_saved();
  }
  last_givens_avoided = givens_avoided;

}/*set_time_saved()---------------------------------------------------*/


void set_change_in_time_saved()
  /*------------------------------------------------------------------
   * Called by semantic_selection()
   * 
   * If a new maxiset has been kept since the last call, reset the
   * estimated change in time_saving accordingly.
   * -----------------------------------------------------------------*/
{
  static int last_kept = 0;
  static int last_time_saved = 0;
  int this_time_saved;

  if (tStats[KEPT] > last_kept) {

    this_time_saved = est_from_data_array(&time_saved);

    set_data_array(&change_in_time_saved, 
		   (this_time_saved - last_time_saved));

    last_time_saved = this_time_saved;
    last_kept = tStats[KEPT];
  }

}/*set_change_in_time_saved()-----------------------------------------*/

void data_list_to_str(char *array_str, struct data_array *array)
  /*------------------------------------------------------------------
   * Called by persist().  
   * 
   * Convert the "list" of the given "array" to a comma delimited
   * string of its values incased in square brackets.  Places the
   * result in the string pointed to by "array_str".
   * -----------------------------------------------------------------*/

{
  int point;
  char tmp[10];

  *array_str = *tmp = '\0';

  if (array->num_points == 0) return;

  snprintf(tmp,10,"[%.2f",array->list[0]); 
  strncat(array_str,tmp,100);
  
  for (point = 0; 
       point < array->num_points && point < NUM_POINTS;
       point++){
    snprintf(tmp,10,", %.2f",array->list[point]); 
    strncat(array_str,tmp,100);
  }
  strncat(array_str,"]",100);
  
}/*data_list_to_str()------------------------------------------------*/


float est_from_data_array(struct data_array *array)
  /*-----------------------------------------------------------------
   * Called by various functions.  Return average from data array
   * ----------------------------------------------------------------*/
{  
  float average, estimate;

  if (array->num_points == 0) 
    abort_scott("Attempting to estimate from data array \"%s\" "
		"which contains no data",array->name);
  if (array->num_points < 2) {
    char list_str[100];

    data_list_to_str(list_str,array);
    message(2,"");
    message(2,"WARNING: estimating from partial data array"
	    "\n         --> \"%s\" has only %d data points",
	    array->name,array->num_points);
    message(2,"");
  }

  average = (array->total/(float)array->num_points);  

  if (array->num_points < NUM_POINTS)
    estimate = (array->list[array->num_points-1] + gradient(array));
  else 
    estimate = (array->list[NUM_POINTS-1] + gradient(array));
    
  return (average + estimate)/2.0;

}/*est_from_data_array()----------------------------------------------*/


float est_cost_per_clause()
{
  return (est_from_data_array(&cost_per_clause));
}

float est_prob_of_success()
{
  return (est_from_data_array(&prob_of_success));
}

float est_change_in_time_saved()
{
  return (est_from_data_array(&change_in_time_saved));
}

float gradient(struct data_array *array)
  /*------------------------------------------------------------------
   * Called by persist().  Returns least squares estimate of the
   * gradient of a line of best fit.
   * -----------------------------------------------------------------*/
{
  int n;
  float x, y, sum_y, sum_x, sum_xx, sum_xy;
  
  x = y = n = 0;
  sum_x = sum_y = sum_xy= sum_xx = 0;
  
  while (n < array->num_points 
	 && n < NUM_POINTS ){
    x = n;
    y = array->list[n];
    sum_y += y;
    sum_x += x;
    sum_xy += (x*y);
    sum_xx += (x*x);
    n++;
  }
  
  return (((sum_x*sum_y)-(n*sum_xy)) /
	  ((sum_x*sum_x)-(n*sum_xx))  );
  
}/*gradient()---------------------------------------------------------*/

float intercept(struct data_array *array)
  /*------------------------------------------------------------------
   * Called by persist().  Returns least squares estimate of the
   * intercept of a line of best fit.
   * -----------------------------------------------------------------*/
{
  int n;
  float x, y, sum_y, sum_x, sum_xx, sum_xy;
  
  x = y = n = 0;
  sum_x = sum_y = sum_xy= sum_xx = 0;
  
  while (n < array->num_points
	 && n < NUM_POINTS){
    x = n;
    y = array->list[n];
    sum_y += y;
    sum_x += x;
    sum_xy += (x*y);
    sum_xx += (x*x);
    n++;
  }
  
  return (((sum_x*sum_xy)-(sum_xx*sum_y)) /
	  ((sum_x*sum_x)-(n*sum_xx))	  );
  
}/*intercept()---------------------------------------------------------*/


boolean strictly_increasing(struct data_array *array)
  /*------------------------------------------------------------------
   * Called by persist().  
   * Returns TRUE iff values in the data array are strictly increasing.
   * -----------------------------------------------------------------*/
{
  int point;

  if (array->num_points < NUM_POINTS) return false;
  
  for (point = 1; point < NUM_POINTS; point++){
    if (array->list[point - 1] >= array->list[point]) return false;
  }

  return true;

}/*strictly_increasing()---------------------------------------------*/


boolean strictly_decreasing(struct data_array *array)
  /*------------------------------------------------------------------
   * Called by persist().  
   * Returns TRUE iff values in the data array are strictly decreasing.
   * -----------------------------------------------------------------*/
{
  int point;
  
  if (array->num_points < NUM_POINTS) return false;
  
  for (point = 1; point < NUM_POINTS; point++){
    if (array->list[point - 1] <= array->list[point]) return false;
  }

  return true;

}/*strictly_decreasing()---------------------------------------------*/

boolean decreasing(struct data_array *array)
  /*------------------------------------------------------------------
   * Called by persist().  
   * Returns TRUE iff values in the data array are strictly increasing.
   * -----------------------------------------------------------------*/
{
  int point;

  if (array->num_points < NUM_POINTS) return false;
  
  for (point = 1; point < NUM_POINTS; point++){
    if (array->list[point - 1] < array->list[point]) return false;
  }

  return true;

}/*decreasing()-------------------------------------------------------*/

void end_development(char *event);

void end_development(char *event)
{
    message(2,"");
    message(2,"\n%s",event);
    message(2,"");
    print_partition("at end of development");
    tStats[MAXISETS_AT_END_DEV] = num_maxisets;
    develop_semantics = recording = false;
}


boolean persist()
  /*------------------------------------------------------------------
   * Called by execute_strategy()
   * 
   * Determined whether the cost of attempting to form a new maxiset
   * is likely to out-weigh the expected return in terms of guidance.
   * -----------------------------------------------------------------*/
{
  int cost_of_attempt, time_left; // in millisecs
  int  accrued_time_saving; // in millisecs
  float expected_return, prob;
  static int fails = 0;
  char array_str[100];

  if (!develop_semantics) return false;

  if (First->first_cl) return false; // clause in semantic queue.
  if (num_maxisets == tParms[MAX_NUM_MAXISETS].val) return false;
  if (num_maxisets * 2 > Stats[SOS_SIZE]) return false;

  // this array must be contain something for the remaining tests
  if (cost_per_clause.num_points < 2) return true;

  cost_of_attempt = (Stats[SOS_SIZE] + Stats[USABLE_SIZE]) 
    * est_from_data_array(&cost_per_clause);
 
  if (Parms[MAX_SECONDS].val > 0)
    time_left = (Parms[MAX_SECONDS].val*1000) - run_time();
  else time_left = MAX_INT;
  
 // buffer to ensure at least a few maxisets have been kept
  if (NUM_POINTS > tStats[ATTEMPTS] 
      && cost_of_attempt*fParms[COST_MARGIN].val < time_left) {
    return true;
  }
  
  // if (num_maxisets < tParms[MIN_NUM_MAXISETS].val) return true;//??

  message(3,"");
  message(3,"%-22s %4.0f secs","Cost of an attempt",
	  cost_of_attempt/(float)1000);
  message(3,"%-22s %4.0f secs", "Time remaining",
	  time_left/(float)1000);

 // too close to time limit? this should be more
  if (cost_of_attempt*fParms[COST_MARGIN].val > time_left) {
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_COST_VS_TIME]++;
    record_event("<time left>");
    end_development("Too close to time limit");
    return false;
  }

  if (strictly_increasing(&time_saved)) {
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_TIME_INC]++;
    record_event("<t-inc>");
    end_development("Timed saved is strictly increasing");
    return false;
  }

  if (strictly_decreasing(&time_saved)) {
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_TIME_DEC]++;
    record_event("<t-dec>");
    end_development("Timed saved is strictly decreasing");
    return false;
  }
  

  // this array must be full before the remaining tests
  if (prob_of_success.num_points < NUM_POINTS) return true;
  
  *array_str = '\0';
  data_list_to_str(array_str,&prob_of_success);
  
  if (get_time_limit()*fParms[TOLERANCE].val > 
      tParms[MAX_SEARCH_TIME].val // at max search time
      && strictly_decreasing(&prob_of_success)) {
    //  if (gradient(prob_of_success) > 0) {
    char tmp[30];
    fails++;
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_PROB_DEC]++;
    record_event("<p-dec>");
    sprintf(tmp,"Probability that attempt will succeed is "
	    "strictly decreasing\n---->%s", array_str);
    end_development(tmp);
    return false;
  }
  
  prob = est_from_data_array(&prob_of_success);
  message(3,"%-22s    %.4f", "Prob of success",prob);
  
  if (cost_of_attempt*(1-prob)*fParms[COST_MARGIN].val > time_left) {
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_COST_VS_PROB]++;
    record_event("<prob too low>");
    end_development("Probability of success too low");
    return false;
  }
  
  if (num_maxisets < tParms[MIN_NUM_MAXISETS].val) return true;

  // this array must be full before the remaining tests
  if (time_saved.num_points < NUM_POINTS) return true;
  
  accrued_time_saving = 
    time_saved.total
    - scott_clock_val(SEM_DEV_TIME) 
    - scott_clock_val(SEM_GUIDE_TIME);
  message(3,"%-22s  %4.0f secs", "Accrued time saving",
	  accrued_time_saving/(float)1000);
  
  // have we accumulated enough time savings for a new attempt?
  if (accrued_time_saving < 
      cost_of_attempt*fParms[COST_MARGIN].val){
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_TIME_ACCRUED]++;
    record_event("<t-acr>");
    message(2,"");
    return false;    
  }

  // this array must be full before the remaining tests
  if (change_in_time_saved.num_points < NUM_POINTS) return true;
  
  expected_return = 
    est_from_data_array(&change_in_time_saved)*prob;
  
  message(3,"%-22s %4.0f secs","Expected return",
	  expected_return/(float)1000);
  
  if (cost_of_attempt*fParms[COST_MARGIN].val > expected_return) {
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_COST_VS_RETURN]++;
    record_event("<return too low>");
    end_development("Expected return too small");
    return false;
  }  
  
  return true;
  
}/*--persist() -------------------------------------------------------*/


