/*------------------------------------------------------------------------*\
                               Sem_res.c

  This code contains the extra functions need to implement semantic
  resolution.  

  DOES NOT WORK YET (DONT KNOW IF IT EVER WILL).

\*------------------------------------------------------------------------*/

#include "SCOTT.h"

#define MAX_SEM_RES_DEPTH 10

void set_up_model()
  /*------------------------------------------------------------------
   * Called by sem_res() 
   * 
   * Attempts to model the given resolvent in the "current_theory"
   * -----------------------------------------------------------------*/
{
  struct clause *cl;
  boolean modelled;

  // check that we can do this?
  
  for (cl=First->first_cl; cl; cl = cl->next_cl) {
    modelled = false;
    if (false_in_model(cl)) {
      printf("\nCould not model INPUT clause %d ",cl->id);
      printf("for constraint set.\n");
    }
  }
  
  for (cl = Usable->first_cl; cl; cl=cl->next_cl) {
    if (false_in_model(cl)) {
      printf("\nCould not model USABLE clause %d ",cl->id);
      printf("for constraint theory.\n");
    } 
  }
  
  for (cl = Sos->first_cl; cl; cl=cl->next_cl) {
    false_in_model(cl);
  }

}/*--set_up_model()---------------------------------------------------*/


boolean false_in_model(struct clause *cl)
  /*------------------------------------------------------------------------*\
    Called by sem_res()
    Attempts to model the given resolvant in the current set
    Returns true iff it fails to do so.
  \*------------------------------------------------------------------------*/
{
  char finder_cl[CLAUSE_STRING_ARRAY];

  if (!cl->first_lit) return false;  // can't model empty clause!
 
  renumber_vars(cl); // normally done in proc

  if (!make_finder_clause(finder_cl, cl)) {
    disable_sem_res();
    return false;
  }

  if (model(finder_cl, constraint_set, false)) {
    cl->in_set[constraint_set] = true;
    return false;
  }
  
  cl->in_set[constraint_set] = false;
  return true;

} /* false_in_model() -----------------------------------------------------*/ 


void sem_res(struct clause *giv_cl)
  /*------------------------------------------------------------------------*\
    Called by infer_and_process()
    
    Draft implementatoin of (ordered) semantic resolution.  Recall any
    clause can be a nucleus, but only clauses false in the model can
    be satelites.  There are two cases.  First the given clause may be
    a nucleus so clash with USABLE clauses that are FALSE in the
    guiding model.  Next, if the given clause is FALSE in the guiding
    model it can be a satelite, and so every Usable clause is a
    potential nucleus.  Recusively calls sem_res_recuse with the
    selected nucleus.

    need to ensure that only literals for the nucleus are resolved
    away.
  \*------------------------------------------------------------------------*/
{

  struct literal *giv_lit, *last_nuc_lit, *new_last_nuc_lit;
  struct term *nuc_atom, *giv_atom;
  struct context *nuc_subst, *giv_subst;
  struct trail *tr;
  struct fpa_tree *ut;
  struct fpa_index *db;
  struct clause *resolvent;
  
  CLOCK_START(SEM_RES_TIME);
  if (!giv_cl->first_lit) {// skip empty clauses
    CLOCK_STOP(SEM_RES_TIME);
    return;
  }

  // either TRUE of FALSE giv_cl can be a nucleus
  last_nuc_lit = giv_cl->first_lit;
  while (last_nuc_lit->next_lit) last_nuc_lit = last_nuc_lit->next_lit;

  sem_res_recurse(giv_cl, true,last_nuc_lit);

  // but if it TRUE it can't be a satelite too
  if (giv_cl->in_set[constraint_set]) {
    CLOCK_STOP(SEM_RES_TIME);
    return;
  }

  giv_subst = get_context();
  giv_subst->multiplier = 0;
  nuc_subst = get_context();
  nuc_subst->multiplier = 1;
  giv_lit = giv_cl->first_lit;
  new_last_nuc_lit = NULL;
    
  // find something that clashes with the giv_cl first
  while (giv_lit != NULL) {
    if (maximal_lit(giv_lit) ) { // ordered sem_res
      giv_atom = giv_lit->atom;
      if (giv_atom->varnum != ANSWER)  {  /* if not answer literal */
	
	if (giv_lit->sign) db = Fpa_clash_neg_lits;
	else db = Fpa_clash_pos_lits;
	
	ut = build_tree(giv_lit->atom, UNIFY,Parms[FPA_LITERALS].val, db);
	
	nuc_atom = next_term(ut, 0);
	while (nuc_atom != NULL) {

	  // find the last literal of the nucleus
	  last_nuc_lit = giv_cl->first_lit;
	  while (last_nuc_lit->next_lit) last_nuc_lit = last_nuc_lit->next_lit;

	  tr = NULL;
	  if (unify(nuc_atom, nuc_subst, giv_atom, giv_subst, &tr)) {
	    resolvent = build_sem_res(nuc_atom,
				      nuc_subst,
				      giv_atom,
				      giv_subst,
				      true,
				      last_nuc_lit,
				      &(new_last_nuc_lit));
	    clear_subst_1(tr);
	    if (false_in_model(resolvent)) {
	      Stats[CL_GENERATED]++;
	      Stats[SEM_RES_GEN]++;
	      CLOCK_STOP(SEM_RES_TIME);
	      pre_process(resolvent, 0, Sos);
	      p_clause(resolvent);
	      CLOCK_START(SEM_RES_TIME);
	    }
	    else {
	      sem_res_recurse(resolvent, false,new_last_nuc_lit); 
	      cl_del_non(resolvent); // free up memory
	    }
	  }
	  
	  nuc_atom = next_term(ut, 0);
	}
      }
    }
    giv_lit = giv_lit->next_lit;
  }
  
  free_context(nuc_subst);
  free_context(giv_subst);
  
  CLOCK_STOP(SEM_RES_TIME);
} /* sem_res() -------------------------------------------------------------*/


void sem_res_recurse(struct clause *nuc, boolean start, struct literal *last_nuc_lit)
  /*------------------------------------------------------------------------*\
    Called by infer_and_process()
    
    Draft implementation of (ordered) semantic resolution.  Recusively
    tries to resolve away literals in "nuc" until a resolvant that is
    false in the guidaing model is produced.
    
    since we always place new literals at the end of a partial resolvant,
  \*------------------------------------------------------------------------*/

{
  struct literal *nuc_lit, *new_last_nuc_lit;
  struct term *nuc_atom, *sat_atom;
  struct context *nuc_subst, *sat_subst;
  struct trail *tr;
  struct fpa_tree *ut;
  struct fpa_index *db;
  struct clause *resolvent;
  
  nuc_subst = get_context();
  nuc_subst->multiplier = 0;
  sat_subst = get_context();
  sat_subst->multiplier = 1;
  nuc_lit = nuc->first_lit;
  new_last_nuc_lit = NULL;

  while (nuc_lit != NULL) {

    nuc_atom = nuc_lit->atom;
    if (nuc_atom->varnum != ANSWER)  {  /* if not answer literal */
      
      if (nuc_lit->sign) db = Fpa_clash_neg_lits;
      else db = Fpa_clash_pos_lits;
      
      ut = build_tree(nuc_lit->atom, UNIFY,
		      Parms[FPA_LITERALS].val, db);
      
      sat_atom = next_term(ut, 0);
      while (sat_atom != NULL) {
	tr = NULL;

	if (!sat_atom->occ.lit->container->in_set[constraint_set] &&
	    maximal_lit(sat_atom->occ.lit) && // ordered sem_res
	    unify(nuc_atom, nuc_subst, sat_atom, sat_subst, &tr)) {
	  
	  resolvent = build_sem_res(nuc_atom, 
				    nuc_subst, 
				    sat_atom, 
				    sat_subst,
				    start,
				    last_nuc_lit,
				    &(new_last_nuc_lit));
	  clear_subst_1(tr);
	  if (false_in_model(resolvent)) {
	    Stats[CL_GENERATED]++;
	    Stats[SEM_RES_GEN]++;
	    CLOCK_STOP(SEM_RES_TIME);
	    pre_process(resolvent, 0, Sos);
	    p_clause(resolvent);
	    CLOCK_START(SEM_RES_TIME);
	  }
	  else {
	    sem_res_recurse(resolvent,false,last_nuc_lit);
	    cl_del_non(resolvent); // free up memory
	  }
	}
	
	sat_atom = next_term(ut, 0);
      }
    }
    if (nuc_lit == last_nuc_lit) break;
    nuc_lit = nuc_lit->next_lit;
  }
  
  free_context(nuc_subst);
  free_context(sat_subst);
  
}  /* sem_res_recurse()-------------------------------------------------- */


struct clause *build_sem_res(struct term *nuc_atom,
			     struct context *nuc_subst,
			     struct term *sat_atom,
			     struct context *sat_subst,
			     boolean start,
			     struct literal *last_nuc_lit,
			     struct literal **new_last_nuc_lit)
  /*-----------------------------------------------------------------------*\
    Called by sem_res() and sem_res_recurse() Just like
    build_bin_res() except history is appeand to the existing nucleus
    history.
  \*------------------------------------------------------------------------*/
{
    struct clause *res;
    struct literal *lit, *new, *prev;
    struct int_ptr *ip1, *ip2, *ip3, *old_ip;
    boolean back_up = false;

    if (last_nuc_lit->atom == nuc_atom) back_up = true; 

    res = get_clause();
    prev = NULL;

    // nucleus literals first
    lit = nuc_atom->occ.lit->container->first_lit;
    while (lit != NULL) {
      if (back_up && lit->next_lit == last_nuc_lit)
	*new_last_nuc_lit = lit;
      if (lit->atom != nuc_atom) {
	new = get_literal();
	new->container = res;
	if (prev == NULL)
	  res->first_lit = new;
	else
	  prev->next_lit = new;
	prev = new;
	new->sign = lit->sign;
	new->atom = apply(lit->atom, nuc_subst);
	new->atom->occ.lit = new;
	new->atom->varnum = lit->atom->varnum;  /* copy type of atom */
      }
      lit = lit->next_lit;
    }
    
    // then the satelites
    lit = sat_atom->occ.lit->container->first_lit;
    while (lit != NULL) {
      if (lit->atom != sat_atom) {
	new = get_literal();
	new->container = res;
	if (res->first_lit == NULL)
	  res->first_lit = new;
	else
	  prev->next_lit = new;
	prev = new;
	new->sign = lit->sign;
	new->atom = apply(lit->atom, sat_subst);
	new->atom->occ.lit = new;
	new->atom->varnum = lit->atom->varnum;  /* copy type of atom */
      }
      lit = lit->next_lit;
    }
    
    
    if (start) {
      // nucleus not just a partial resolvant
      ip3 = get_int_ptr();
      ip1 = get_int_ptr();
      ip2 = get_int_ptr();
      ip1->i = nuc_atom->occ.lit->container->id;
      ip2->i = sat_atom->occ.lit->container->id;
      ip3->i = SEM_RES_RULE;
      ip3->next = ip1;
      ip1->next = ip2;
      res->parents = ip3;
    }
    else {
      // make a copy of the existing history of the nucleus
      ip1 = get_int_ptr();
      res->parents = ip1;

      old_ip = nuc_atom->occ.lit->container->parents;
      ip1->i = old_ip->i;
      old_ip = old_ip->next;
      while (old_ip) {
	ip2 = get_int_ptr();
	ip2 ->i = old_ip->i;
	ip1->next = ip2;
	ip1 = ip2;
	old_ip = old_ip->next;
      }

      // add the satelite
      ip2 = get_int_ptr();
      ip2->i = sat_atom->occ.lit->container->id;
      ip1->next = ip2;
    }

    // no detailed history yet

    if (Flags[PROG_SYNTHESIS].val)
	combine_answers(res, nuc_atom, nuc_subst, sat_atom, sat_subst);

    return(res);

}  /* build_sem_res()----------------------------------------------------*/


void disable_sem_res()
  /*------------------------------------------------------------------------*\
    Called by false_in_model() Clause is too large too model so we
    diable the sem_res inference rule.  If using autonomous mode we
    perform a syntactic analysis in order to set soem sensible
    inference rules.
  \*------------------------------------------------------------------------*/
{

  struct clause *c;
  static boolean first_call = true;
  int propositional = false;
  int equality = false;
  int symmetry = false;
  int horn = false;
  int max_lits = 0;
  int i;

  if (!first_call) return; // in case we're in the middle of sem_res
  first_call = false;

  printf("\n\nDisabling sem_res inference rule. ");

  auto_change_flag(stdout, SEM_RES, 0);

  // sem_res only set for no-prop clause spaces (??)
  
  /* Find out some basic properties. */

  for (c=Usable->first_cl, horn = 1; c && horn; c=c->next_cl)
    horn = horn_clause(c);
  for (c=Sos->first_cl; c && horn; c=c->next_cl)
    horn = horn_clause(c);
  for (c=First->first_cl; c && horn; c=c->next_cl)
    horn = horn_clause(c);

  for (c=Usable->first_cl, equality = 0; c && !equality; c=c->next_cl)
    equality = equality_clause(c);
  for (c=Sos->first_cl; c && !equality; c=c->next_cl)
    equality = equality_clause(c);
  for (c=First->first_cl; c && !equality; c=c->next_cl)
    equality = equality_clause(c);

  for (c=Usable->first_cl, max_lits = 0; c; c=c->next_cl) {
    i = num_literals(c);
    max_lits = (i > max_lits ? i : max_lits);
  }
  
  for (c=Sos->first_cl; c; c=c->next_cl) {
    i = num_literals(c);
    max_lits = (i > max_lits ? i : max_lits);
  }
  for (c=First->first_cl; c; c=c->next_cl) {
    i = num_literals(c);
    max_lits = (i > max_lits ? i : max_lits);
  }

  printf("\nProperties of clause space: prop=%d, horn=%d, equality=%d, symmetry=%d, max_lits=%d.\n\n",
	   propositional, horn, equality, symmetry, max_lits);
	
  /* assumes nonpropositional */
  
  if (max_lits == 1) {
    printf("Setting pick_given_ratio to 2, ");
    printf("because all clauses are units.\n");
    auto_change_parm(stdout, PICK_GIVEN_RATIO, 2);
  }
  else {
    /* Nonunit */
    printf("Setting hyper_res, because there are nonunits.\n");
    auto_change_flag(stdout, HYPER_RES, 1);
    if (equality || !horn) {
      printf("Setting ur_res, because this is a nonunit set containing\n");
      printf("either equality literals or non-Horn clauses.\n");
      auto_change_flag(stdout, UR_RES, 1);
    }
    
    if (horn) {
      printf("Clearing order_hyper, because all clauses are Horn.\n");
      auto_change_flag(stdout, ORDER_HYPER, 0);
    }
    else {
      /* Non-Horn */
      printf("Setting factor and  unit_deletion, because there are non-Horn clauses.\n");
      auto_change_flag(stdout, FACTOR, 1);
      auto_change_flag(stdout, UNIT_DELETION, 1);
    }
  }

} /* disable_sem_res() -----------------------------------------------------*/
