/*------------------------------------------------------------------
 * Functions for managing parameters local to SCOTT sources.
 *-----------------------------------------------------------------*/

#include "SCOTT.h"

void init_scott_options()
  /*------------------------------------------------------------------
   * Called by OTTER fcn init().  Initialise options local to SCOTT
   * sources.
   * -----------------------------------------------------------------*/
{
  int parm;
  
  for (parm = 0; parm < MAX_TPARMS; parm++)    
    tParms[parm].name = "";
  
  for (parm = 0; parm < MAX_FPARMS; parm++)
    fParms[parm].name = "";

  tParms[CONTROL_MEMORY_LIMIT].name = "control_memory_limit";
  tParms[CONTROL_MEMORY_LIMIT].val = 0; // soft limit
  tParms[CONTROL_MEMORY_LIMIT].min = 0;
  tParms[CONTROL_MEMORY_LIMIT].max = MAX_INT;

  tParms[LIMIT_WEIGHT_AT].name = "limit_weight_at";
  tParms[LIMIT_WEIGHT_AT].val = 10; // percentile
  tParms[LIMIT_WEIGHT_AT].min = 0;
  tParms[LIMIT_WEIGHT_AT].max = 100;
  
  tParms[CULL_WEIGHTS_OVER].name = "cull_weights_over";
  tParms[CULL_WEIGHTS_OVER].val = 90;  // percentile
  tParms[CULL_WEIGHTS_OVER].min = 0;
  tParms[CULL_WEIGHTS_OVER].max = 100;

  tParms[LIMIT_VARS_AT].name = "limit_vars_at";
  tParms[LIMIT_VARS_AT].val = 50; // percentile
  tParms[LIMIT_VARS_AT].min = 0;
  tParms[LIMIT_VARS_AT].max = 100;
  
  tParms[CULL_VARS_OVER].name = "cull_vars_over";
  tParms[CULL_VARS_OVER].val = 90;  // percentile
  tParms[CULL_VARS_OVER].min = 0;
  tParms[CULL_VARS_OVER].max = 100;

  tParms[MAX_DOMAIN_SIZE].name = "max_domain_size";
  tParms[MAX_DOMAIN_SIZE].val = 10;
  tParms[MAX_DOMAIN_SIZE].min = 1;
  tParms[MAX_DOMAIN_SIZE].max = MAX_INT;

  tParms[MIN_DOMAIN_SIZE].name = "min_domain_size";
  tParms[MIN_DOMAIN_SIZE].val = 0;
  tParms[MIN_DOMAIN_SIZE].min = 0;
  tParms[MIN_DOMAIN_SIZE].max = MAX_INT;

  tParms[MAX_NUM_MAXISETS].name ="max_num_maxisets";
  tParms[MAX_NUM_MAXISETS].val = 5;
  tParms[MAX_NUM_MAXISETS].min = 0;
  tParms[MAX_NUM_MAXISETS].max = MAXISET_ARRAY-2;
  
  tParms[MIN_NUM_MAXISETS].name ="min_num_maxisets";
  tParms[MIN_NUM_MAXISETS].val = 3;
  tParms[MIN_NUM_MAXISETS].min = 0;
  tParms[MIN_NUM_MAXISETS].max = MAXISET_ARRAY-2;
  
  tParms[MIN_MAXISET_ATTEMPTS].name ="min_maxiset_attempts";
  tParms[MIN_MAXISET_ATTEMPTS].val = 5;
  tParms[MIN_MAXISET_ATTEMPTS].min = 1;
  tParms[MIN_MAXISET_ATTEMPTS].max = MAXISET_ARRAY-2;
  
  tParms[MAX_SEARCH_TIME].name = "max_search_time";
  tParms[MAX_SEARCH_TIME].val = 2000;  // millisecs
  tParms[MAX_SEARCH_TIME].min = 0;
  tParms[MAX_SEARCH_TIME].max = MAX_INT;
  
  tParms[MIN_SEARCH_TIME].name = "min_search_time";
  tParms[MIN_SEARCH_TIME].val = 500;   // millisecs
  tParms[MIN_SEARCH_TIME].min = 0;
  tParms[MIN_SEARCH_TIME].max = MAX_INT;
  
  tParms[TRUE_WEIGHT].name ="true_weight";
  tParms[TRUE_WEIGHT].val = 0;
  tParms[TRUE_WEIGHT].min = -MAX_INT;
  tParms[TRUE_WEIGHT].max = MAX_INT;
  
  tParms[VERBOSITY].name = "verbosity";
  tParms[VERBOSITY].val = 1;
  tParms[VERBOSITY].min = 0;
  tParms[VERBOSITY].max = 5;
  
  tParms[MAX_NEW_BASE_CL].name = "max_new_base_cl";
  tParms[MAX_NEW_BASE_CL].val = 5;
  tParms[MAX_NEW_BASE_CL].min = 0;
  tParms[MAX_NEW_BASE_CL].max = 10;
  
  tParms[SEM_QUEUE_LIMIT].name = "sem_queue_limit";
  tParms[SEM_QUEUE_LIMIT].val = 2;
  tParms[SEM_QUEUE_LIMIT].min = 0;
  tParms[SEM_QUEUE_LIMIT].max = MAX_INT;
  
  tParms[MIN_REFINEMENT].name = "min_refinement";
  tParms[MIN_REFINEMENT].val = 0;
  tParms[MIN_REFINEMENT].min = -MAX_INT;
  tParms[MIN_REFINEMENT].max = 0;
  
  tParms[MAX_FAILS].name ="max_fails";
  tParms[MAX_FAILS].val = 2;
  tParms[MAX_FAILS].min = 0;
  tParms[MAX_FAILS].max = MAX_INT;
  
  tParms[MAX_NULL_SET_SIZE].name ="max_null_set_size";
  tParms[MAX_NULL_SET_SIZE].val = 5;
  tParms[MAX_NULL_SET_SIZE].min = 0;
  tParms[MAX_NULL_SET_SIZE].max = MAX_INT;
  
  // float parameters
  fParms[MIN_PROB_EXTN].name = "min_prob_extn";
  fParms[MIN_PROB_EXTN].val = 0.3;
  fParms[MIN_PROB_EXTN].min = 0;
  fParms[MIN_PROB_EXTN].max = 1;
  
  fParms[SCALE_WEIGHT].name ="scale_weight";
  fParms[SCALE_WEIGHT].val = 0;
  fParms[SCALE_WEIGHT].min = 0;
  fParms[SCALE_WEIGHT].max = MAX_INT;
  
  fParms[TOLERANCE].name ="tolerance";
  fParms[TOLERANCE].val = 1.5;
  fParms[TOLERANCE].min = 1;
  fParms[TOLERANCE].max = MAX_INT;
  
  fParms[SMOOTHING_RATIO].name ="smoothing_ratio";
  fParms[SMOOTHING_RATIO].val = 0.225;
  fParms[SMOOTHING_RATIO].min = 0;
  fParms[SMOOTHING_RATIO].max = 1;
  
  fParms[GUIDANCE_FACTOR].name ="guidance_factor";
  fParms[GUIDANCE_FACTOR].val = 5;
  fParms[GUIDANCE_FACTOR].min = 0;
  fParms[GUIDANCE_FACTOR].max = MAX_INT;
  
  fParms[COST_MARGIN].name ="cost_margin";
  fParms[COST_MARGIN].val = 10;
  fParms[COST_MARGIN].min = 0;
  fParms[COST_MARGIN].max = MAX_INT;
  
  fParms[MIN_COST_OF_ATTEMPT].name ="min_cost_of_attempt";
  fParms[MIN_COST_OF_ATTEMPT].val = 2.0;  // seconds
  fParms[MIN_COST_OF_ATTEMPT].min = 0.0;
  fParms[MIN_COST_OF_ATTEMPT].max = MAX_INT;
  
}/* init_scott_options() ---------------------------------------------*/

void auto_change_parm_str(FILE *fp, char *parm_name, float new_value)
{
  int index;
  
  // integer parm?
  for (index = 0; tParms[index].name && index < MAX_TPARMS; index++) {
    if (str_ident(parm_name, tParms[index].name)){ 
      tParms[index].val = (int) new_value;
      fprintf(fp, "   dependent: assign(%s,%.0f).\n",parm_name,new_value);
      return;
    }
  }
  // float parm?
  for (index = 0; fParms[index].name && index < MAX_FPARMS; index++) {
    if (str_ident(parm_name, fParms[index].name)) {
      fParms[index].val = new_value;
      fprintf(fp, "   dependent: assign(%s,\"%.2f\").\n",parm_name,new_value);
      return;
    }  
  }
}/*auto_change_parm_str()---------------------------------------------*/


int change_scott_parm(FILE *fp, struct term *t, char *parm_name)
  /*------------------------------------------------------------------
   * Called by OTTER fcn change_parm().  Checks for parameter from
   * amongst those local to SCOTT sources. If found, returns its id,
   * else returns -1.
   * -----------------------------------------------------------------*/
{
  int index, rc;
  char *int_name;

  for (index = 0; 
       tParms[index].name && index < MAX_TPARMS; 
       index++) {
    int new_val;
    
    if (str_ident(parm_name, tParms[index].name)) {
      
      int_name = sn_to_str(t->farg->narg->argval->sym_num);
      rc = str_int(int_name, &new_val);
      if (rc == 0) {
	fprintf(fp, "ERROR: ");
	print_term(fp, t);
	fprintf(fp, " second argument must be an integer.\n");
	Stats[INPUT_ERRORS]++;
	return(-1);
      }
      
      if (new_val < tParms[index].min || 
	  new_val > tParms[index].max) {
	fprintf(fp, "ERROR: ");
	print_term(fp, t);
	fprintf(fp, " float must be in range [%d,%d].\n",
		tParms[index].min, tParms[index].max);
	Stats[INPUT_ERRORS]++;
	return(-1);
      }
      
      if (new_val == tParms[index].val) {
	fprintf(fp, "WARNING: ");
	print_term(fp, t);
	fprintf(fp, " already has that value.\n");
	return(index);
      }
      tParms[index].val = new_val;
      
      return(index);
    }
  }

  for (index = 0; 
       fParms[index].name && index < MAX_FPARMS; 
       index++) {
    double new_fval;
    
    if (str_ident(parm_name, fParms[index].name)) {
      
      int_name = sn_to_str(t->farg->narg->argval->sym_num);
      rc = str_double(int_name, &new_fval);
      if (rc == 0) {
	fprintf(fp, "ERROR: ");
	print_term(fp, t);
	fprintf(fp, " second argument must be a float.\n");
	Stats[INPUT_ERRORS]++;
	return(-1);
      }
      
      if (new_fval < fParms[index].min || 
	  new_fval > fParms[index].max) {
	fprintf(fp, "ERROR: ");
	print_term(fp, t);
	fprintf(fp, " float must be in range [%.2f,%.2f].\n",
		fParms[index].min, fParms[index].max);
	Stats[INPUT_ERRORS]++;
	return(-1);
      }
      
      if (new_fval == fParms[index].val) {
	fprintf(fp, "WARNING: ");
	print_term(fp, t);
	fprintf(fp, " already has that value.\n");
	return(index);
      }
      fParms[index].val = new_fval;
      
      return(index);
    }
  }

  return -1;

}/*change_scott_parm()------------------------------------------------*/



