/*--prototypes in SCOTT.c--*/

int semantic_wt(struct clause *cl);
void add_semantic_wt(struct clause *cl);
void retest_failed_clauses(int maxiset);
void remodel_failed_sos_clauses(int maxiset);
void check_failed_clauses();
struct clause *semantic_selection(struct list *lptr);
struct clause *find_best_sem_cl(struct list *lptr, int set_to_skip);
void extend_base(int key_maxiset);
boolean execute_strategy(boolean (*strategy)());
void update_partition(int new_maxiset);
void reset_partition(int bad_maxiset);
boolean extend_conjecture(int new_maxiset);
boolean keep(int new_maxiset);
boolean discard_bad_maxisets();
void print_in_set(struct external_list *elist);
void shuffle_ext_list(struct external_list *elist);
void scan_new_base_clauses(int new_maxiset);
void scan_conjecture(int new_maxiset);
boolean start_scan(int new_maxiset);
void scan_base(int new_maxiset);
void reset_base_markers();
void scan_sos(int new_maxiset);
void scan_usable(int new_maxiset);
void scan_first(int new_maxiset);
boolean model_clause(struct clause *cl, int maxiset);
boolean test_clause(struct clause *cl, int maxiset);
boolean verify_model_in_list(int maxiset, struct list *lptr, char list_name[]);
void verify_model(int maxiset);
void discard_subsets(int new_maxiset);
void discard(int bad_maxiset);
void disable_semantics(char *event);
void init_semantics(int argc, char *argv[]);
void output_finder_stats(FILE *fp, int stats_level);
long scott_clock_val(int c);
void message(int priority, char *msg, ...);
void abort_scott(char *msg, ...);
void record_event(char *event, ...);
void output_history();
int get_num_maxisets();
boolean in_constraint_set(struct clause *cl);


/*--prototypes in Sauto.c--*/

boolean all_clauses(int (*property)());
boolean at_least_one_clause(int (*property)());
int get_max(int (*property)());
int get_min(int (*property)());
int get_num(int (*property)());
int mix_clause(struct clause *c);
int clause_weight(struct clause *c);
int pos_ground_clause(struct clause *c);
int neg_ground_clause(struct clause *c);
int equality_axiom(struct clause *c);
void automatic_3_settings();
void auto_set_sos();
void auto_set_conjecture();
void set_properties();
void check_eq_settings();


/*--prototypes in Sclause_space.c--*/

void init_clause_dptr();
struct clause_dptr *get_clause_dptr();
void free_clause_dptr(struct clause_dptr *p);
void delete_cl_ext(struct clause *c, struct external_list *elist);
void prepend_cl_ext(struct clause *c, struct external_list *elist);
void append_cl_ext(struct clause *c, struct external_list *elist);
void sos_scan_insert(struct clause *c);
void sos_scan_delete(struct clause *c);
void declare_soft_equality();
void declare_hard_equality();
void set_base();
void init_clause_space();
boolean make_finder_clause(char *finder_clause, struct clause *otter_clause);
boolean add_finder_term(char *finder_clause, struct term *otter_term);
void in_set_to_str(char *in_set, struct clause *cl);
boolean in_set_identical(struct clause *c1, struct clause *c2);
void print_sos_scan();
void print_failed_clause_nos();
void print_conjecture_nos();
void print_base_nos();
void print_ext_list(struct external_list *elist);
void debug_scan(char *event);
void print_sos();
void print_base();
void print_new_base();
void print_conjecture();
void print_usable();
void print_first();


/*--prototypes in Scontrol.c--*/

int clause_pick_weight(struct clause *cl);
void init_distributions();
void init_a_distribution(struct Distn *distn);
void get_distn(struct Distn *distn);
void set_distn(struct Distn *distn);
void output_distn(struct Distn *distn);
void output_weight_distn();
void output_variable_distn();
int percentile(struct Distn *distn, int percent);
void cull_clauses_over(int max_value, int (*property)());
int weight_percentile(int percent);
int variable_percentile(int percent);
void scott_control_memory();
void control_distn(struct Distn *distn);


/*--prototypes in Scost.c--*/

void init_costs();
void reset_prob_of_success();
void set_data_array(struct data_array *array, float value);
void set_cost_of_given(float value);
void set_cost_per_clause(float value);
void set_prob_of_success(float value);
void set_time_saved(int givens_avoided);
void set_change_in_time_saved();
void data_list_to_str(char *array_str, struct data_array *array);
float est_from_data_array(struct data_array *array);
float est_cost_per_clause();
float est_prob_of_success();
float est_change_in_time_saved();
float gradient(struct data_array *array);
float intercept(struct data_array *array);
boolean strictly_increasing(struct data_array *array);
boolean strictly_decreasing(struct data_array *array);
boolean decreasing(struct data_array *array);
void end_development(char *event);

void end_development(char *event);
boolean persist();


/*--prototypes in Sdeclare_symbols.c--*/

boolean insert_predicate(int sym_num);
boolean insert_function(int sym_num);
boolean insert_constant(int sym_num);
boolean scan_term_for_symbols(struct term *subterm);
boolean scan_list_for_symbols(struct list *list_ptr);
boolean scan_lists_for_symbols();
void debug_print_symbols();
boolean declare_otter_symbols();


/*--prototypes in Sem_res.c--*/

void set_up_model();
boolean false_in_model(struct clause *cl);
void sem_res(struct clause *giv_cl);
void sem_res_recurse(struct clause *nuc, boolean start, struct literal *last_nuc_lit);
struct clause *build_sem_res(struct term *nuc_atom,
			     struct context *nuc_subst,
			     struct term *sat_atom,
			     struct context *sat_subst,
			     boolean start,
			     struct literal *last_nuc_lit,
			     struct literal **new_last_nuc_lit);
void disable_sem_res();


/*--prototypes in Soptions.c--*/

void init_scott_options();
void auto_change_parm_str(FILE *fp, char *parm_name, float new_value);
int change_scott_parm(FILE *fp, struct term *t, char *parm_name);


/*--prototypes in Spartition.c--*/

int get_sos_partition(int maxiset);
int get_size_on_sos(int maxiset);
int get_size_on_usable(int maxiset);
void print_a_partition(struct partition *part_ptr);
void print_partition(char *event);
boolean subset(int set1, int set2);
boolean sos_subset(int set1, int set2);
boolean already_exists(int new_maxiset);
int superset_of(int new_maxiset);
void check_for_model(int new_maxiset);
float size_choice_space();
float rate_partition();
float rate_maxiset(int skip_set);
float rate_clause(struct clause *cl);
void set_sem_queue(struct list *lptr);
int next_set_to_skip();
int variance(struct clause *cl);
boolean cover_new_clauses(int key);
boolean refine_choice(int key);
boolean in_key_set(int key_set, struct clause *cl);
struct clause *search_list_1(struct list *lptr, int skip_maxiset);
struct clause *search_list_2(struct list *lptr, int skip_maxiset);
struct clause *search_list_3(struct list *lptr, int skip_maxiset);
boolean model_usable();
void init_a_partition(struct partition *part_ptr);
void init_partitions();
void init_sizes(int maxiset);
void usable_part_insert_all(struct clause *cl);
void sos_part_insert_all(struct clause *cl);
void usable_part_delete_all(struct clause *cl);
void sos_part_delete_all(struct clause *cl);
void scott_insert(struct list *lptr, struct clause *cl);
void scott_delete(struct list *lptr, struct clause *cl);
void scott_delete_from_conjecture(struct clause *cl);
void usable_part_insert(struct clause *cl, int new_maxiset);
void sos_part_insert(struct clause *cl, int new_maxiset);
void usable_part_delete(struct clause *cl, int old_maxiset);
void sos_part_delete(struct clause *cl, int old_maxiset);
void sos_part_mv_to_set(struct clause *cl, int old_maxiset);
void usable_part_mv_to_set(struct clause *cl, int old_maxiset);
void init_partition_box(int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY]);
int get_list_part(char *list_name, struct list *lptr,
		  int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		  int cmp_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY]);
int compare_partitions(char *list_name,
		       int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		       int set_in_list_cell_copy[MAXISET_ARRAY][MAXISET_ARRAY],
		       int cmp_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		       int cmp_in_list_cell_copy[MAXISET_ARRAY][MAXISET_ARRAY]);
void debug_partition(char *event);


